open Camlp4.PreCast
open Syntax

let header = "
#define raise(e) caml_raise_constant( *caml_named_value(e))
#define matrix(m, n1, n2, d) int n1 = Bigarray_val(m)->dim[0];\
                             int n2 = Bigarray_val(m)->dim[1];\
                             d = Data_bigarray_val(m)
#define vector(v, n, d) int n = Bigarray_val(v)->dim[0];\
                        d = Data_bigarray_val(v)
#define identity(x) x
"

(* from ocaml source code *)
let create_hashtable size init =
  let tbl = Hashtbl.create size in
	List.iter (fun (key, data) -> Hashtbl.add tbl key data) init;
	tbl

let fromval = create_hashtable 31 [
  "int"   , ("Int_val"   , "int");
  "string", ("String_val", "char*");
  "bool"  , ("Bool_val"  , "char");
  "float" , ("Double_val", "double");
]

let toval = create_hashtable 31 [
  "unit"  , "Val_unit";
  "int"   , "Val_int";
  "string", "copy_string";
  "bool"  , "Val_bool";
  "float" , "copy_double";
]

(* convert a type to a string reprensenting it *)
let rec string_of_ctyp = function
  | Ast.TyArr (_loc, t1, t2) -> (string_of_ctyp t1)^" -> "^(string_of_ctyp t2)
  | (Ast.TyId (_, Ast.IdLid (_, s ))) -> s
  | _ -> failwith "error in string_of_ctyp"

(* build the name of the c source code file using the location information *)
let cname_of_loc loc = (Filename.chop_extension (Camlp4.PreCast.Syntax.Ast.Loc.file_name loc))^"_ciml.c"

(* the temporary C file in which the C source is written *)
let output_ref = ref None

(* open the c file and put the initial definitions if necessary, or
   simply return the file descriptor *)
let output loc =
  match !output_ref with
    | Some fd -> fd
    | None ->
        let out = open_out (cname_of_loc loc) in
        let inc s =
          Printf.fprintf out "#include <%s>\n%!" s
        in
          output_ref := Some out;
          inc "caml/mlvalues.h";
          inc "caml/alloc.h";
          inc "caml/memory.h";
          inc "caml/fail.h";
          inc "caml/callback.h";
          inc "caml/custom.h";
          inc "caml/intext.h";
          inc "caml/bigarray.h";
          Printf.fprintf out "%s%!" header;
          out

(* Contains the last block of C code *)
let c_code = ref ""

(* Directly ouput C code to the temp file if it is in str_item position *)
let cstm loc _ s = Printf.fprintf (output loc) "%s\n%!" s; <:str_item@loc<  >>

(* if the code is in expr position, then store it in the c_code buffer *)
let cfun loc _ s = c_code := s; <:expr@loc< >>

(* register a new handler for conversion to caml values *)
let rec_toval loc a t = Hashtbl.add toval (string_of_ctyp t) a; <:str_item@loc<  >>

(* unregister an handler for conversion to caml values *)
let unrec_toval loc t = Hashtbl.remove toval (string_of_ctyp t); <:str_item@loc<  >>

(* register a new handler for conversion from caml values *)
let rec_fromval loc a b t = Hashtbl.add fromval (string_of_ctyp t) (a,b); <:str_item@loc<  >>

(* unregister an handler for conversion from caml values *)
let unrec_fromval loc t = Hashtbl.remove fromval (string_of_ctyp t); <:str_item@loc<  >>

let warning s =
  Printf.fprintf stderr "Warning: %s" s

(* generate the C function and the corresponding external declaration*)
let gen_fun _loc u args t =
  (* create the type of the function, takes a list of (string * ctype) option *)
  let rec mk_type acc = function
      []   -> acc
    | None::q ->
        let acc = (<:ctyp< unit -> $acc$ >>)
        in mk_type acc q
    | (Some (a,t))::q ->
        let acc = (<:ctyp< $t$ -> $acc$ >>)
        in mk_type acc q
  in
  (* utility function to flatten a list of string into a string using "," as
   * separator *)
  let comma s e = if s = "" then s^e else s^", "^e in
  (* build the caml to c conversion *)
  let declare_conv uc = function
	| None -> ""
	| Some (v,t) ->
		let st = string_of_ctyp t in
		  try
			let f,t = Hashtbl.find fromval st in
			  Printf.sprintf "%s %s = %s(%s__%i);\n" t v f v uc
		  with
			  Not_found -> warning (Printf.sprintf "conversion handler fromval not found for type %s\n%!" st);
				Printf.sprintf "value %s = %s__%i;\n" v v uc
  in
  (* put the #define for the type conversion of the return value from c to caml *)
  let define_return t =
	let st = string_of_ctyp t in
	  if st = "unit" then
		Printf.sprintf "#define Return(foo) CAMLreturn(Val_unit)\n"
	  else
		try
		  let f = Hashtbl.find toval st in
			Printf.sprintf "#define Return(foo) CAMLreturn(%s(foo))\n" f
		with
			Not_found ->  warning (Printf.sprintf "conversion handler toval not found for type %s\n%!" st);
			  Printf.sprintf "#define Return(foo) CAMLreturn(foo)\n"
  in
  (* cancel the type conversion of the return value *)
  let undef_return () =
	Printf.sprintf "#undef Return\n"
  in
  (* retrieve the body of the function *)
  let c_body = !c_code in
  (* number of parameters *)
  let nb_params = List.length args in
  Printf.fprintf (output _loc) "%s" (
    Printf.sprintf "value %s (%s) {\n  %s\n%s\n%s%s\n%s}\n\n%!"
      (* function name *)
      u
      (* arguments *)
      (fst (List.fold_left (fun (acc, uc) n ->
        match n with
          None -> (comma acc "value unit__"^(string_of_int uc), uc+1)
        | Some (a,t) -> (comma acc "value "^a^"__"^(string_of_int uc), uc+1)
      ) ("", 0) args))
      (* the CAMLparam declaration *)
      (Printf.sprintf "CAMLparam%d(%s);"
        nb_params
        (fst (List.fold_left (fun (acc, uc) n ->
          match n with
            None -> (comma acc "unit__"^(string_of_int uc), uc+1)
          | Some (a,t) -> (comma acc (a^"__"^(string_of_int uc)), uc+1)
        ) ("", 0) args))
      )
	  (* type conversions for the arguments *)
      (fst (List.fold_left (fun (acc, uc) vt -> (acc^(declare_conv uc vt)), uc+1) ("",0) args))
	  (* type conversion for the return value *)
	  (define_return t)
      (* the body of the function *)
      c_body
	  (* cleaning of the type conversion for the return value *)
	  (undef_return ())
  );
  (* return the external declaration *)
  <:str_item< external $lid:u$ : $mk_type t (List.rev args)$ = $Ast.LCons (u, Ast.LNil)$ >>

EXTEND Gram
  str_item:
[
  [
	"letext"; u = a_LIDENT; args = LIST1
	  [
		"("; ")" -> None
	  | "("; a = LIDENT; ":"; t = ctyp; ")" -> Some (a,t)
	  ]; ":"; t = ctyp; "="; e = expr -> gen_fun _loc u args t
  ]
| [
	"register_toval"; a = STRING; ":"; t = ctyp -> rec_toval _loc a t
  ]
| [
	"unregister_toval"; ":"; t = ctyp -> unrec_toval _loc t
  ]
| [
	"register_fromval"; a = STRING; b = STRING; ":"; t = ctyp -> rec_fromval _loc a b t
  ]
| [
	"unregister_fromval"; ":"; t = ctyp -> unrec_fromval _loc t
  ]
];
END;;

let _ =
  Quotation.add "" Quotation.DynAst.str_item_tag cstm;
  Quotation.add "" Quotation.DynAst.expr_tag cfun

